UNIT frmMain;
{+F
---------------------------------------------------------------------------
  File        - FRMMAIN.PAS
  Purpose     - Demonstrate MD5
  Author      - Keith S. Brown
  Revised     - 2000.07.03 (KSB) Wrote initial version.
---------------------------------------------------------------------------}
INTERFACE
USES
  Windows,       Messages,      SysUtils,      Classes,       Graphics,
  Controls,      Forms,         Dialogs,       StdCtrls,      Buttons,
  ExtCtrls;

TYPE
  TForm1           = CLASS(TForm)
    btnDoIt        : TSpeedButton;
    btnExit        : TSpeedButton;
    btnFile        : TRadioButton;
    btnSelect      : TSpeedButton;
    btnText        : TRadioButton;
    bvlResult      : TBevel;
    dlgOpen        : TOpenDialog;
    edtFileName    : TEdit;
    edtText        : TEdit;
    lblDigest      : TLabel;
    pnlButton      : TPanel;

    procedure   Do_ClickTarget(aoSender: TObject);
    PROCEDURE   Do_Exit(aoSender: TObject);
    PROCEDURE   Do_FileSelect(aoSender: TObject);
    PROCEDURE   Do_MD5(aoSender: TObject);
  END {CLASS};

VAR
  Form1: TForm1;
              {=================================================}
IMPLEMENTATION
USES
  pcEncode;

{$R *.DFM}


{}PROCEDURE   TForm1.Do_ClickTarget(aoSender: TObject);
{+M
---------------------------------------------------------------------------
  Purpose     - EVENT HANDLER. Toggle radio buttons appropriately.
---------------------------------------------------------------------------}
  BEGIN
    IF aoSender IS TEdit THEN BEGIN
      IF TEdit(aoSender).Name = edtText.Name THEN BEGIN
        btnText.Checked := TRUE;
      END ELSE
      IF TEdit(aoSender).Name = edtFileName.Name THEN BEGIN
        btnFile.Checked := TRUE;
      END {IF};
    END {IF};
{}END {Do_ClickTarget};




{}PROCEDURE TForm1.Do_Exit(aoSender: TObject);
{+M
---------------------------------------------------------------------------
  Purpose     - EVENT HANDLER. Exit application.
---------------------------------------------------------------------------}
  BEGIN
    Close;
{}END {Do_Exit};




{}PROCEDURE   TForm1.Do_MD5(aoSender: TObject);
{+M
---------------------------------------------------------------------------
  Purpose     - EVENT HANDLER. Execute MD5.
---------------------------------------------------------------------------}
  VAR
    md5hash   : TCrMD5;                               (* Step 1a *)
  BEGIN
    md5hash := TCrMD5.Create;                         (* Step 1b *)
    TRY
      IF btnText.Checked THEN BEGIN
        md5hash.InputType    := hsString;             (* Step 2 *)
        md5hash.InputString  := edtText.Text;         (* Step 3 *)
        md5hash.InputLength  := Length(edtText.Text); (* Step 4 *)
        md5hash.MD5_Hash;                             (* Step 5 *)
        lblDigest.Caption    := md5hash.DigestHex;
      END ELSE
      IF btnFile.Checked THEN BEGIN
        md5hash.InputType    := hsFile;               (* Step 2 *)
        md5hash.InputFilePath:= edtFileName.Text;     (* Step 3 *)
        md5hash.MD5_Hash;                             (* Step 4 *)
        lblDigest.Caption    := md5hash.DigestHex;
      END ELSE BEGIN
        ShowMessage('Do what!');
      END {IF};

    FINALLY
      md5hash.Free;
    END {TRY};
{}END {Do_MD5};




{}PROCEDURE   TForm1.Do_FileSelect(aoSender: TObject);
{+M
---------------------------------------------------------------------------
  Purpose     - EVENT HANDLER. Select file to process.
---------------------------------------------------------------------------}
  BEGIN
    IF dlgOpen.Execute THEN BEGIN
      edtFileName.Text := dlgOpen.FileName;
    END {IF};
{}END {Do_FileSelect};




END {frmMain}.
