UNIT pcEncode;
{+F
---------------------------------------------------------------------------
  File        - PCENCODE.PAS
  Purpose     - Provides MD5 & CRC functions.
  Description - This unit contains an Object Pascal Class which can be used to
                perform an MD5 Hashing of byte array, file, or Pascal String.
                A MD5 Hashing or Message Digest is a 'finger print' of the
                input. This is 100% PASCAL!!!

                "It is conjectured that it is computationally infeasible
                to produce two messages having the same message digest"....
                "The MD5 algorithm is intended for digital signature
                applications, where a large file must be "compressed" in a
                secure manner before being encrypted with a private (secret)
                key under a public-key cryptosystem such as RSA." R. Rivest
                RfC: 1321, RSA Data Security, Inc. April 1992

                The MD5 Algorithm was produced by RSA Data Security Inc.
                (See LEGAL)

  Author      - Grg Carter; gregc@cryptocard.com
  Organization- CRYPTOCard Corporation, info@cryptocard.com
                R&D Division, Carleton Place, ON, CANADA, K7C 3T2
                1-613-253-3152 Voice, 1-613-253-4685 Fax.

  Copyright    1996 by Cryptocard.

  LEGAL       - The algorithm was placed into the public domain, hence requires
                no license or runtime fees.  However this code is copyright by
                CRYPTOCard. CRYPTOCard grants anyone who may wish to use, modify
                or redistribute this code privileges to do so, provided the user
                agrees to the following three(3) rules:

                1)Any Applications, (ie exes which make use of this
                  Object...), for-profit or non-profit, must acknowledge the
                  author of this Object(ie. MD5 Implementation provided by Greg
                  Carter, CRYPTOCard Corporation) somewhere in the accompanying
                  Application documentation(ie AboutBox, HelpFile, readme...).
                  NO runtime or licensing fees are required!

                2)Any Developer Component(ie Delphi Component, Visual Basic VBX,
                  DLL) derived from this software must acknowledge that it is
                  derived from "MD5 Object Pascal Implementation Originated by
                  Greg Carter, CRYPTOCard Corporation 1996". Also all efforts
                  should be made to point out any changes from the original.
                  !!!!!Further, any Developer Components based on this code
                  *MAY NOT* be sold for profit.  This Object was placed into the
                  public domain, and therefore any derived components should
                  also.!!!!!

                3)CRYPTOCard Corporation makes no representations concerning
                  this software or the suitability of this software for any
                  particular purpose. It is provided "as is" without express or
                  implied warranty of any kind. CRYPTOCard accepts no liability
                  from any loss or damage as a result of using this software.

                CRYPTOCard Corporation is in no way affiliated with RSA Data
                Security Inc.
                The MD5 Algorithm was produced by RSA Data Security Inc.

  Usage       - Below is typical usage(for File)of the MD5 Object, Follow these
                steps:
        Step 1: Declare and Create a New TCrMD5 object.  This can be done by
                'Drag N Drop' a TCrMD5 off the Delphi Tool Pallet,
                or explicitly in code.
        Step 2: Set the InputType.
        Step 3: Point to the input(hsString, hsFile, hsByteArray).
        Step 4: Call the MD5_Hash procedure.
        Step 5: Read the Digest from DigestHex, DigestStr or Digest.
                You are Done!
  Example     - (MD5 of a file)
---------------------------------------------------------------------------
procedure Tcryptfrm.Button1Click(Sender: TObject);
var
  loMD5       : TCrMD5;                            (* Step 1a *)
begin
  loMD5 := TCrMD5.Create;                          (* Step 1b *)
  try
    If OpenDialog1.Execute then begin
      loMD5.InputType := hsFile;                   (* Step 2 *)
      loMD5.InputFilePath := OpenDialog1.FileName; (* Step 3 *)
      loMD5.MD5_Hash;                              (* Step 4 *)
      Label2.Caption := loMD5.DigestHex;           (* Do something with output *)
    end;
  finally
    loMD5.free;
  end;
end;
---------------------------------------------------------------------------
  Why Use this instead of a freely available C DLL?
                The goal was to provide a number of Encryption/Hash
                implementations in Object Pascal, so that the Pascal Developer
                has considerably more freedom.  These Implementations are geared
                toward the PC(Intel) Microsoft Windows developer, who will be
                using Borland's New 32bit developement environment(Delphi32).
                The code generated by this new compiler is considerablely
                faster then 16bit versions. And should provide the Developer
                with faster implementations then those using C DLLs.

  NOTES       - Version 1 does not contain any cross-platform considerations.
                If trying to use this code on a Big Endian style processor you
                will need to write additional code to reorder the bytes.

  Revised     - 1996.01.03 (GC) Version 1.
              - Compatibility & Testing with BP7.0: Anne Marcel Roorda, garfield@xs4all.nl
              - 2000.07.03 (KSB) Renamed SourceType to THashSource and enumerated values
                from (InputString, InputFilePath, InputArray) to (hsFile, hsByteArray, hsString).
                Reformatted code for readability. Added DigestHex & DigestStr properties.
---------------------------------------------------------------------------}
{$A+   Word align variables }
{$F+   Force Far calls }
{$K+   Use smart callbacks
{$N+   Allow coprocessor instructions }
{$P+   Open parameters enabled }
{$S+   Stack checking }
{$T-   @ operator is NOT typed }
{$U-   Non Pentium safe FDIV }
{$Z-   No automatic word-sized enumerations}
{$H+   Huge Strings}

{$RANGECHECKS OFF}
{$OVERFLOWCHECKS OFF}
INTERFACE
USES
  SysUtils, Classes,  Windows;


TYPE
  THashSource = (hsFile, hsByteArray, hsString);

TYPE
  TBuffer     = ARRAY[0..4159]OF BYTE;
  ULONG32     = RECORD
    LoWord16  : WORD;
    HiWord16  : WORD;
  END {RECORD};

  PULONG32    = ^ULONG32;
  PLong       = ^LongINT;

  TDigest     = ARRAY[0..15]OF BYTE;
  THashDigest = RECORD
    A : DWORD;
    B : DWORD;
    C : DWORD;
    D : DWORD;
  END {RECORD};

  PTR_Hash    = ^THashDigest;

  TCrMD5      = CLASS
  PRIVATE
    FaDigest       : TDigest;
    FeSource       : THashSource;      {Source type, whether its a file or ByteArray, or a Pascal String}
    FsInputFilePath: STRING;           {Full Path to Input File}
    FpInputBytes   : PByte;            {Point to input array}
    FsInput        : STRING;           {Input String}
    FpDigest       : PTR_Hash;         {output MD5 Digest}
    FiSourceLength : LongINT;          {input length in Bytes}
    FaActiveBlock  : ARRAY[0..15] OF DWORD; {the 64Byte block being transformed}

    FA             : DWORD;       {FA..FDD are used during Step 4, the transform.}
    FB             : DWORD;       {I made them part of the Object to cut down on}
    FC             : DWORD;       {time used to pass variables.}
    FD             : DWORD;
    FAA            : DWORD;
    FBB            : DWORD;
    FCC            : DWORD;
    FDD            : DWORD;
              {FF, GG, HH, II are used in Step 4, the transform}
    PROCEDURE   FF(VAR a, b, c, d, x : DWORD; s : BYTE; ac : DWORD);
    PROCEDURE   GG(VAR a, b, c, d, x : DWORD; s : BYTE; ac : DWORD);
    PROCEDURE   HH(VAR a, b, c, d, x : DWORD; s : BYTE; ac : DWORD);
    PROCEDURE   II(VAR a, b, c, d, x : DWORD; s : BYTE; ac : DWORD);
  PROTECTED
    FUNCTION    MD5_GetDigestHex:STRING;
    FUNCTION    MD5_GetDigestStr:STRING;
    PROCEDURE   MD5_Initialize;   {Initialize is used in Step 3, this fills FA..FD with init. values}
    PROCEDURE   MD5_Transform;    {this is where all the magic happens}
    PROCEDURE   MD5_Finish;
    PROCEDURE   MD5_Hash_Block(aaBuffer:TBuffer; aiBytesInBlock:LongINT);
    PROCEDURE   MD5_Hash_Bytes;
    PROCEDURE   MD5_Hash_File;
  PUBLIC
    PROCEDURE   MD5_Hash;         {This procedure sends the data 64Bytes at a time to MD5_Transform}
  PUBLIC
    PROPERTY    pInputArray  : PByte       READ FpInputBytes    WRITE FpInputBytes;
    PROPERTY    pOutputArray : PTR_Hash    READ FpDigest        WRITE FpDigest; {!!See FOutputArray}
    PROPERTY    Digest       : TDigest     READ FaDigest;
    PROPERTY    DigestHex    : STRING      READ MD5_GetDigestHex;
    PROPERTY    DigestStr    : STRING      READ MD5_GetDigestStr;
  PUBLISHED
    PROPERTY    InputType    : THashSource READ FeSource        WRITE FeSource;
    PROPERTY    InputFilePath: STRING      READ FsInputFilePath WRITE FsInputFilePath;
    PROPERTY    InputString  : STRING      READ FsInput         WRITE FsInput;
    PROPERTY    InputLength  : LongINT     READ FiSourceLength  WRITE FiSourceLength;
  END {CLASS};

// CRC 32 Base Functions

{}FUNCTION    CRC32(aiCRC:LongINT; CONST aiByte:BYTE):LongINT;
{}FUNCTION    CRCStr(asText : STRING) : LongINT;
              {=================================================}
IMPLEMENTATION
CONST
{Constants for MD5Transform routine.}
  S11 = 7;
  S12 = 12;
  S13 = 17;
  S14 = 22;
  S21 = 5;
  S22 = 9;
  S23 = 14;
  S24 = 20;
  S31 = 4;
  S32 = 11;
  S33 = 16;
  S34 = 23;
  S41 = 6;
  S42 = 10;
  S43 = 15;
  S44 = 21;
CONST
  CRC32_POLYNOMIAL = $EDB88320;

VAR
  uaCcitt32Table   : ARRAY[0..255] OF LongINT;



{}FUNCTION    CrcStr(asText : STRING) : LongINT;
{+H
---------------------------------------------------------------------------
  Purpose     - Return the CRC value of a string.
---------------------------------------------------------------------------}
  VAR
    li        : INTEGER;
  BEGIN
    result := 0;
    FOR li := 1 TO Length(asText) DO BEGIN
      result := CRC32(result, BYTE(asText[li]));
    END {FOR};
{}END {CrcStr};




{}FUNCTION    CRC32(aiCRC:LongINT; CONST aiByte:BYTE):LongINT;
{+H
---------------------------------------------------------------------------
  Purpose     - Return the CRC value of a Byte.
---------------------------------------------------------------------------}
  BEGIN
    result := (((aiCRC SHR 8) AND $00FFFFFF) XOR (uaCcitt32Table[(aiCRC XOR aiByte) AND $FF]));
{}END {CRC32};




{}PROCEDURE   BuildCRCTable;
{+H
---------------------------------------------------------------------------
  Purpose     - Build a 32 bit CRC lookup table.
---------------------------------------------------------------------------}
  VAR
    li        : DWORD;
    liJ       : DWORD;
    liValue   : DWORD;
  BEGIN
    FOR li := 0 TO 255 DO BEGIN
      liValue := li;
      FOR liJ := 8 DOWNTO 1 DO BEGIN
        IF ((liValue AND 1) <> 0) THEN BEGIN
          liValue := (liValue SHR 1) XOR CRC32_POLYNOMIAL
        END ELSE BEGIN
          liValue := liValue SHR 1;
        END {IF};
      END {FOR};

      uaCcitt32Table[li] := liValue;
    END {FOR};
{}END {BuildCRCtable};


{MD5 Unit}
{This will only work on an intel}

CONST
  ciMaxBlock  = 4096;


{}FUNCTION    DigestToHex(aaDigest:TDigest):STRING;
{+H
---------------------------------------------------------------------------
  Purpose     - Return a MD5 digest as a string of hexidecimal characters.
---------------------------------------------------------------------------}
  VAR
    li        : INTEGER;
  BEGIN
    result := '';
    FOR li := 0 TO 15 DO BEGIN
      result := result + IntToHex(aaDigest[li],2);
    END {FOR};
{}END {DigestToHex};




{$WARNINGS off}
{}FUNCTION    ROL(A : LongINT; Amount : BYTE) : LongINT;
{+H
---------------------------------------------------------------------------
  Purpose     - Rotate Left.
---------------------------------------------------------------------------}
  BEGIN
    ASM
      mov cl, Amount
      mov eax, a
      rol eax, cl
      mov result, eax
    END {ASM};
{}END {ROL};
{$WARNINGS on}




{}PROCEDURE   TCrMD5.FF;
{+M
---------------------------------------------------------------------------
  Purpose     - Round 1 of the Transform.
                Equivalent to a = b + ((a + F(b,c,d) + x + ac) <<< s)
                Where F(b,c,d) = b And c Or Not(b) And d
---------------------------------------------------------------------------}
  BEGIN
    a := a + ((b AND c) OR (NOT (b) AND (d))) + x + ac;
    a := ROL(a, s);
    Inc(a, b);
{}END {FF};




{}PROCEDURE TCrMD5.GG;
{+M
---------------------------------------------------------------------------
  Purpose     - Round 2 of the Transform.
                Equivalent to a = b + ((a + G(b,c,d) + x + ac) <<< s)
                Where G(b,c,d) = b And d Or c Not d
---------------------------------------------------------------------------}
  BEGIN
    a := a + ((b AND d) OR (c AND (NOT d))) + x + ac;
    a := ROL(a, s);
    Inc(a, b);
{}END {GG};




{}PROCEDURE   TCrMD5.HH;
{+M
---------------------------------------------------------------------------
  Purpose     - Round 3 of the Transform.
                Equivalent to a = b + ((a + H(b,c,d) + x + ac) <<< s)
                Where H(b,c,d) = b Xor c Xor d
---------------------------------------------------------------------------}
  BEGIN
    a := a + (b XOR c XOR d) + x + ac;
    a := ROL(a, s);
    a := b + a;
{}END {HH};




{}PROCEDURE TCrMD5.II;
{+M
---------------------------------------------------------------------------
  Purpose     - Round 4 of the Transform.
                Equivalent to a = b + ((a + I(b,c,d) + x + ac) <<< s)
                Where I(b,c,d) = C Xor (b Or Not(d))
---------------------------------------------------------------------------}
  BEGIN
    a := a + (c XOR (b OR (NOT d))) + x + ac;
    a := ROL(a, s);
    a := b + a;
{}END {II};




{}FUNCTION    TCrMD5.MD5_GetDigestHex:STRING;
{+M
---------------------------------------------------------------------------
  Purpose     - Return the MD5 digest as a 32 character hex string.
---------------------------------------------------------------------------}
  BEGIN
    result := DigestToHex(Digest);
{}END {MD5_GetDigestHex};




{}FUNCTION    TCrMD5.MD5_GetDigestStr:STRING;
{+M
---------------------------------------------------------------------------
  Purpose     - Return the MD5 digest as a 16 character ascii string.
---------------------------------------------------------------------------}
  VAR
    li        : INTEGER;
  BEGIN
    result := '';
    FOR li := 0 TO 15 DO BEGIN
      result := result + CHAR(Digest[li]);
    END {FOR};
{}END {MD5_GetDigestStr};




{}PROCEDURE   TCrMD5.MD5_Initialize;
{+M
---------------------------------------------------------------------------
  Purpose     - Initialize the MD5 hash.
---------------------------------------------------------------------------}
  BEGIN
    FA := $67452301;
    FB := $EFCDAB89;
    FC := $98BADCFE;
    FD := $10325476;
{}END {MD5_Initialize};




{}PROCEDURE   TCrMD5.MD5_Hash;
{+M
---------------------------------------------------------------------------
  Purpose     - Hash the input value.
---------------------------------------------------------------------------}
  VAR
    lpStr     : PChar;
  BEGIN
    MD5_Initialize;
    CASE FeSource OF
      hsFile : BEGIN
        MD5_Hash_File;
      END; {hsFile}

      hsByteArray : BEGIN
        MD5_Hash_Bytes;
      END; {hsByteArray}

      hsString : BEGIN
              {Convert Pascal String to Byte Array}
        lpStr := NIL;
        TRY   {protect dyanmic memory allocation}
          GetMem(lpStr, Length(FsInput) + 1);
          StrPCopy(lpStr, FsInput);
          FiSourceLength := Length(FsInput);
          FpInputBytes   := Pointer(lpStr);
          MD5_Hash_Bytes;
        FINALLY
          IF lpStr <> NIL THEN BEGIN
            FreeMem(lpStr, Length(FsInput) + 1);
          END {IF};
        END {TRY};
      END {hsString};
    END {CASE};

    MD5_Finish;
{}END {MD5_Hash};




{}PROCEDURE   TCrMD5.MD5_Hash_Block(aaBuffer:TBuffer; aiBytesInBlock:LongINT);
{+M
---------------------------------------------------------------------------
  Purpose     - Hash a buffer of bytes.
---------------------------------------------------------------------------}
  VAR
    liIndex   : INTEGER;
  BEGIN
    liIndex := 0;
    REPEAT
      Move(aaBuffer[liIndex], FaActiveBlock, 64);
      MD5_Transform;
      Inc(liIndex,64);
    UNTIL (liIndex = aiBytesInBlock);
{}END {MD5_Hash_Block};




{}PROCEDURE   TCrMD5.MD5_Hash_Bytes;
{+M
---------------------------------------------------------------------------
  Purpose     - Hash a byte array.
  Revised     - 2000.07.04.03 (KSB) Rewrote to use all bytes, not just the
                first 4160 bytes.
---------------------------------------------------------------------------}
  VAR
    laBuffer       : TBuffer;
    lbFinished     : BOOLEAN;
    liBytesRead    : INTEGER;
    liCount64      : Comp;
    liIndex        : LongINT;
  BEGIN
    lbFinished:= FALSE;
    liCount64 := 0;
    liIndex   := 0;
    REPEAT
              {extract bytes from FpInputBytes in blocks of 4096.  If
               the remaining unprocessed bytes is less than 4096 than
               take all that remain}
      IF FiSourceLength > liIndex THEN BEGIN
        liBytesRead := FiSourceLength - liIndex;
        IF liBytesRead > ciMaxBlock THEN BEGIN
          liBytesRead := ciMaxBlock;
        END {IF};

        Move(PByteArray(FpInputBytes)^[liIndex], laBuffer, liBytesRead);
        Inc(liIndex,ciMaxBlock);

        liCount64 := liCount64 + liBytesRead;
        IF liBytesRead <> ciMaxBlock THEN BEGIN
          laBuffer[liBytesRead] := $80;
          Inc(liBytesRead);
          WHILE (liBytesRead MOD 64) <> 56 DO BEGIN
            laBuffer[liBytesRead] := 0;
            Inc(liBytesRead);
          END {WHILE};

          liCount64 := liCount64 * 8;
          Move(liCount64, laBuffer[liBytesRead],8);
          Inc(liBytesRead,8);
          lbFinished := TRUE;
        END {IF};

        MD5_Hash_Block(laBuffer,liBytesRead);
      END ELSE BEGIN
        Break;
      END {IF};
    UNTIL lbFinished;
{}END {Hash_Bytes};




{}PROCEDURE   TCrMD5.MD5_Hash_File;
{+M
---------------------------------------------------------------------------
  Purpose     - Hash a file.
---------------------------------------------------------------------------}
  VAR
    laBuffer       : TBuffer;
    lbFinished     : BOOLEAN;
    liBytesRead    : INTEGER;
    liCount64      : Comp;
    lFile          : FILE;
  BEGIN
    AssignFile(lFile, FsInputFilePath);
    Reset(lFile, 1);
    lbFinished := FALSE;
    liCount64  := 0;
    REPEAT
      BlockRead(lFile, laBuffer, 4096, liBytesRead);
      liCount64 := liCount64 + liBytesRead;
      IF liBytesRead <> 4096 {reached end of file} THEN BEGIN
        laBuffer[liBytesRead] := $80;
        Inc(liBytesRead);
        WHILE (liBytesRead MOD 64) <> 56 DO BEGIN
          laBuffer[liBytesRead] := 0;
          Inc(liBytesRead);
        END {WHILE};

        liCount64 := liCount64 * 8;
        Move(liCount64, laBuffer[liBytesRead], 8);
        Inc(liBytesRead, 8);
        lbFinished := True;
      END {IF};

      MD5_Hash_Block(laBuffer,liBytesRead);
    UNTIL lbFinished;

    CloseFile(lFile);
{}END {MD5_Hash_File};




{}PROCEDURE   TCrMD5.MD5_Finish;
{+M
---------------------------------------------------------------------------
  Purpose     - Finish the MD5 hash.
---------------------------------------------------------------------------}
  BEGIN
    IF FpDigest <> NIL THEN BEGIN
      FpDigest^.A := FA;
      FpDigest^.B := FB;
      FpDigest^.C := FC;
      FpDigest^.D := FD;
    END {IF};

    THashDigest(FaDigest).A := FA;
    THashDigest(FaDigest).B := FB;
    THashDigest(FaDigest).C := FC;
    THashDigest(FaDigest).D := FD;
{}END {MD5_Finish};




{}PROCEDURE   TCrMD5.MD5_Transform;
{+M
---------------------------------------------------------------------------
  Purpose     - Perform Step 4 of the algorithm.  This is where all the
                important stuff happens.  This performs the rounds on a
                64Byte Block.  This procedure should be called in a loop
                until all input data has been transformed.
---------------------------------------------------------------------------}
  BEGIN
    FAA := FA;
    FBB := FB;
    FCC := FC;
    FDD := FD;

              { Round 1 }
    FF(FA, FB, FC, FD, FaActiveBlock[00], S11, $D76AA478); { 1 }
    FF(FD, FA, FB, FC, FaActiveBlock[01], S12, $E8C7B756); { 2 }
    FF(FC, FD, FA, FB, FaActiveBlock[02], S13, $242070DB); { 3 }
    FF(FB, FC, FD, FA, FaActiveBlock[03], S14, $C1BDCEEE); { 4 }
    FF(FA, FB, FC, FD, FaActiveBlock[04], S11, $F57C0FAF); { 5 }
    FF(FD, FA, FB, FC, FaActiveBlock[05], S12, $4787C62A); { 6 }
    FF(FC, FD, FA, FB, FaActiveBlock[06], S13, $A8304613); { 7 }
    FF(FB, FC, FD, FA, FaActiveBlock[07], S14, $FD469501); { 8 }
    FF(FA, FB, FC, FD, FaActiveBlock[08], S11, $698098D8); { 9 }
    FF(FD, FA, FB, FC, FaActiveBlock[09], S12, $8B44F7AF); { 10 }
    FF(FC, FD, FA, FB, FaActiveBlock[10], S13, $FFFF5BB1); { 11 }
    FF(FB, FC, FD, FA, FaActiveBlock[11], S14, $895CD7BE); { 12 }
    FF(FA, FB, FC, FD, FaActiveBlock[12], S11, $6B901122); { 13 }
    FF(FD, FA, FB, FC, FaActiveBlock[13], S12, $FD987193); { 14 }
    FF(FC, FD, FA, FB, FaActiveBlock[14], S13, $A679438E); { 15 }
    FF(FB, FC, FD, FA, FaActiveBlock[15], S14, $49B40821); { 16 }

              { Round 2 }
    GG(FA, FB, FC, FD, FaActiveBlock[01], S21, $F61E2562); { 17 }
    GG(FD, FA, FB, FC, FaActiveBlock[06], S22, $C040B340); { 18 }
    GG(FC, FD, FA, FB, FaActiveBlock[11], S23, $265E5A51); { 19 }
    GG(FB, FC, FD, FA, FaActiveBlock[00], S24, $E9B6C7AA); { 20 }
    GG(FA, FB, FC, FD, FaActiveBlock[05], S21, $D62F105D); { 21 }
    GG(FD, FA, FB, FC, FaActiveBlock[10], S22,  $2441453); { 22 }
    GG(FC, FD, FA, FB, FaActiveBlock[15], S23, $D8A1E681); { 23 }
    GG(FB, FC, FD, FA, FaActiveBlock[04], S24, $E7D3FBC8); { 24 }
    GG(FA, FB, FC, FD, FaActiveBlock[09], S21, $21E1CDE6); { 25 }
    GG(FD, FA, FB, FC, FaActiveBlock[14], S22, $C33707D6); { 26 }
    GG(FC, FD, FA, FB, FaActiveBlock[03], S23, $F4D50D87); { 27 }
    GG(FB, FC, FD, FA, FaActiveBlock[08], S24, $455A14ED); { 28 }
    GG(FA, FB, FC, FD, FaActiveBlock[13], S21, $A9E3E905); { 29 }
    GG(FD, FA, FB, FC, FaActiveBlock[02], S22, $FCEFA3F8); { 30 }
    GG(FC, FD, FA, FB, FaActiveBlock[07], S23, $676F02D9); { 31 }
    GG(FB, FC, FD, FA, FaActiveBlock[12], S24, $8D2A4C8A); { 32 }

              { Round 3 }
    HH(FA, FB, FC, FD, FaActiveBlock[05], S31, $FFFA3942); { 33 }
    HH(FD, FA, FB, FC, FaActiveBlock[08], S32, $8771F681); { 34 }
    HH(FC, FD, FA, FB, FaActiveBlock[11], S33, $6D9D6122); { 35 }
    HH(FB, FC, FD, FA, FaActiveBlock[14], S34, $FDE5380C); { 36 }
    HH(FA, FB, FC, FD, FaActiveBlock[01], S31, $A4BEEA44); { 37 }
    HH(FD, FA, FB, FC, FaActiveBlock[04], S32, $4BDECFA9); { 38 }
    HH(FC, FD, FA, FB, FaActiveBlock[07], S33, $F6BB4B60); { 39 }
    HH(FB, FC, FD, FA, FaActiveBlock[10], S34, $BEBFBC70); { 40 }
    HH(FA, FB, FC, FD, FaActiveBlock[13], S31, $289B7EC6); { 41 }
    HH(FD, FA, FB, FC, FaActiveBlock[00], S32, $EAA127FA); { 42 }
    HH(FC, FD, FA, FB, FaActiveBlock[03], S33, $D4EF3085); { 43 }
    HH(FB, FC, FD, FA, FaActiveBlock[06], S34,  $4881D05); { 44 }
    HH(FA, FB, FC, FD, FaActiveBlock[09], S31, $D9D4D039); { 45 }
    HH(FD, FA, FB, FC, FaActiveBlock[12], S32, $E6DB99E5); { 46 }
    HH(FC, FD, FA, FB, FaActiveBlock[15], S33, $1FA27CF8); { 47 }
    HH(FB, FC, FD, FA, FaActiveBlock[02], S34, $C4AC5665); { 48 }

              { Round 4 }
    II(FA, FB, FC, FD, FaActiveBlock[00], S41, $F4292244); { 49 }
    II(FD, FA, FB, FC, FaActiveBlock[07], S42, $432AFF97); { 50 }
    II(FC, FD, FA, FB, FaActiveBlock[14], S43, $AB9423A7); { 51 }
    II(FB, FC, FD, FA, FaActiveBlock[05], S44, $FC93A039); { 52 }
    II(FA, FB, FC, FD, FaActiveBlock[12], S41, $655B59C3); { 53 }
    II(FD, FA, FB, FC, FaActiveBlock[03], S42, $8F0CCC92); { 54 }
    II(FC, FD, FA, FB, FaActiveBlock[10], S43, $FFEFF47D); { 55 }
    II(FB, FC, FD, FA, FaActiveBlock[01], S44, $85845DD1); { 56 }
    II(FA, FB, FC, FD, FaActiveBlock[08], S41, $6FA87E4F); { 57 }
    II(FD, FA, FB, FC, FaActiveBlock[15], S42, $FE2CE6E0); { 58 }
    II(FC, FD, FA, FB, FaActiveBlock[06], S43, $A3014314); { 59 }
    II(FB, FC, FD, FA, FaActiveBlock[13], S44, $4E0811A1); { 60 }
    II(FA, FB, FC, FD, FaActiveBlock[04], S41, $F7537E82); { 61 }
    II(FD, FA, FB, FC, FaActiveBlock[11], S42, $BD3AF235); { 62 }
    II(FC, FD, FA, FB, FaActiveBlock[02], S43, $2AD7D2BB); { 63 }
    II(FB, FC, FD, FA, FaActiveBlock[09], S44, $EB86D391); { 64 }

    Inc(FA, FAA);
    Inc(FB, FBB);
    Inc(FC, FCC);
    Inc(FD, FDD);
              { Zeroize sensitive information}
    FillChar(FaActiveBlock, SizeOf(FaActiveBlock), #0);
{}END {MD5_Transform};




INITIALIZATION
  BuildCRCTable;
END {pcEncode}.
